@@ -146,13 +146,30 @@ def executable_path_or_raise(tool):
 def executable_path(tool):
     return shutil.which(tool)
 
-def link(source, link_name, link_type="hard"):
-    if link_type.lower() == "soft":
-        os.symlink(source, link_name)
-    elif link_type.lower() == "hard":
-        os.link(source, link_name)
-    else:
-        shutil.copyfile(source, link_name)
+def link(source, link_name, link_type='auto'):
+    _logger.debug("Linking source %s to target %s with link_type %s", source, link_name, link_type)
+    if link_type.lower() == 'soft':
+        return os.symlink(source, link_name)
+    elif link_type.lower() == 'hard':
+        return os.link(source, link_name)
+    elif link_type.lower() == 'copy':
+        return shutil.copyfile(source, link_name)
+    elif link_type.lower() == 'auto':
+        try:
+            return link(source=source, link_name=link_name, link_type='hard')
+        except OSError as e:
+            if e.errno != 18:
+                raise e
+        try:
+            return link(source=source, link_name=link_name, link_type='soft')
+        except OSError as e:
+            pass
+
+        return link(source=source, link_name=link_name, link_type='copy')
+
+    raise Exception('Unable to link source {} to target {} with link_type {}', source, link_name, link_type)
+
+
 
 def m2_path():
     return os.getenv("M2_REPO", (pathlib.Path.home() / '.m2').absolute())
@@ -170,7 +187,10 @@ def launch_java(
         jvm_args,
         main_class,
         *app_args,
-        additional_jars=[]
+        additional_jars=[],
+        stdout=None,
+        stderr=None,
+        **subprocess_run_kwargs,
         ):
     java = executable_path('java')
     if not java:
@@ -179,7 +199,7 @@ def launch_java(
     cp = classpath_separator().join([os.path.join(jar_dir, '*')] + additional_jars)
     _logger.debug("class path: %s", cp)
     jvm_args = tuple(arg for arg in jvm_args) if jvm_args else tuple()
-    return subprocess.run((java, '-cp', cp) + jvm_args + (main_class,) + app_args)
+    return subprocess.run((java, '-cp', cp) + jvm_args + (main_class,) + app_args, stdout=stdout, stderr=stderr, **subprocess_run_kwargs)
 
 def run_and_combine_outputs(command, *args):
     return subprocess.check_output((command,) + args, stderr=subprocess.STDOUT)
@@ -194,7 +214,7 @@ def find_endpoint(argv, shortcuts={}):
             indices.append(index)
     return -1 if len(indices) == 0 else indices[-1]
 
-def jgo_main(argv=sys.argv[1:]):
+def jgo_parser():
 
     epilog='''
 The endpoint should have one of the following formats:
@@ -213,7 +233,7 @@ and it will be auto-completed.
 
     parser = argparse.ArgumentParser(
         description     = 'Run Java main class from maven coordinates.',
-        usage           = '%(prog)s [-v] [-u] [-U] [-m] [--ignore-jgorc] [--additional-jars jar [jar ...]] [--additional-endpoints endpoint [endpoint ...]] [JVM_OPTIONS [JVM_OPTIONS ...]] <endpoint> [main-args]',
+        usage           = '%(prog)s [-v] [-u] [-U] [-m] [--ignore-jgorc] [--link-type type] [--additional-jars jar [jar ...]] [--additional-endpoints endpoint [endpoint ...]] [JVM_OPTIONS [JVM_OPTIONS ...]] <endpoint> [main-args]',
         epilog          = epilog,
         formatter_class = argparse.RawTextHelpFormatter
     )
@@ -225,19 +245,27 @@ and it will be auto-completed.
     parser.add_argument('-a', '--additional-jars', nargs='+', help='Add additional jars to classpath', default=[], required=False)
     parser.add_argument( '--additional-endpoints', nargs='+', help='Add additional endpoints', default=[], required=False)
     parser.add_argument('--ignore-jgorc', action='store_true', help='Ignore ~/.jgorc')
+    parser.add_argument('--link-type', default=None, type=str, help='How to link from local maven repository into jgo cache. Defaults to the `links\' setting in ~/.jrunrc or \'auto\' if not specified.', choices=('hard', 'soft', 'copy', 'auto'))
 
+    return parser
+
+def jgo_main(argv=sys.argv[1:], stdout=None, stderr=None):
+
+    parser = jgo_parser()
 
     try:
-        run(parser, argv=argv)
+        completed_process = run(parser, argv=argv, stdout=stdout, stderr=stderr)
+        completed_process.check_returncode()
     except subprocess.CalledProcessError as e:
-        _logger.error("Error in %s: %s", e.cmd, e)
+        _logger.error("Error in `%s': %d", ' '.join(e.cmd), e.returncode)
+        _logger.debug("Exception: %s", e)
         _logger.debug("Debug Trace:", exc_info=True)
         if e.stdout:
-            _logger.debug("\tMaven std out:")
+            _logger.debug("\tstd out:")
             for l in str(e.stdout).split('\\n'):
                 _logger.debug('\t\t%s', l)
         if e.stderr:
-            _logger.debug("Maven std err:")
+            _logger.debug("\tstd err:")
             for l in str(e.stderr).split('\\n'):
                 _logger.debug('\t\t%s', l)
         parser.print_help(file=sys.stderr)
@@ -270,7 +298,7 @@ def default_config():
     config.add_section('settings')
     config.set('settings', 'm2Repo', os.path.join(str(pathlib.Path.home()), '.m2', 'repository'))
     config.set('settings', 'cacheDir', os.path.join(str(pathlib.Path.home()), '.jgo'))
-    config.set('settings', 'links', 'hard')
+    config.set('settings', 'links', 'auto')
 
     # repositories
     config.add_section('repositories')
@@ -344,6 +372,7 @@ def resolve_dependencies(
         endpoint_string,
         cache_dir,
         m2_repo,
+        link_type='auto',
         update_cache=False,
         force_update=False,
         manage_dependencies=False,
@@ -362,10 +391,12 @@ def resolve_dependencies(
     workspace           = workspace_dir_from_coordinates(coordinates, cache_dir=cache_dir)
 
     update_cache = True if force_update else update_cache
-    update_cache = update_cache or os.path.isdir(workspace)
+    update_cache = update_cache \
+                   or not os.path.isdir(workspace) \
+                   or os.path.isdir(workspace) and not os.path.isfile(os.path.join(workspace, 'mainClass'))
 
     if not update_cache:
-        primary_endpoint, workspace
+        return primary_endpoint, workspace
 
     if update_cache:
         shutil.rmtree(workspace, True)
@@ -457,14 +488,14 @@ def resolve_dependencies(
             relevant_jars.append(jar_file_in_workspace)
 
             try:
-                link(os.path.join(m2_repo, *g.split('.'), a, version, jar_file), jar_file_in_workspace)
+                link(os.path.join(m2_repo, *g.split('.'), a, version, jar_file), jar_file_in_workspace, link_type=link_type)
             except FileExistsError as e:
                 # Do not throw exceptionif target file exists.
                 pass
     return primary_endpoint, workspace
 
 
-def run(parser, argv=sys.argv[1:]):
+def run(parser, argv=sys.argv[1:], stdout=None, stderr=None):
 
     config = default_config()
     if not '--ignore-jgorc' in argv:
@@ -491,8 +522,12 @@ def run(parser, argv=sys.argv[1:]):
     if args.verbose > 0:
         _logger.setLevel(logging.DEBUG)
 
+    if args.link_type is not None:
+        config.set('settings', 'links', args.link_type)
+
     cache_dir = settings.get('cacheDir')
     m2_repo   = settings.get('m2Repo')
+    link_type = settings.get('links')
     for repository in args.repository:
         repositories[repository.split('=')[0]] = repository.split('=')[1]
 
@@ -514,15 +549,15 @@ def run(parser, argv=sys.argv[1:]):
         manage_dependencies = args.manage_dependencies,
         repositories        = repositories,
         shortcuts           = shortcuts,
-        verbose             = args.verbose)
+        verbose             = args.verbose,
+        link_type           = link_type)
 
     main_class_file = os.path.join(workspace, primary_endpoint.main_class, 'mainClass') if primary_endpoint.main_class else os.path.join(workspace, 'mainClass')
 
     try:
         with open(main_class_file, 'r') as f:
             main_class = f.readline()
-        launch_java(workspace, jvm_args, main_class, *program_args, additional_jars=args.additional_jars)
-        return
+        return launch_java(workspace, jvm_args, main_class, *program_args, additional_jars=args.additional_jars, stdout=stdout, stderr=stderr, check=False)
     except FileNotFoundError:
         pass
 
@@ -548,8 +583,7 @@ def run(parser, argv=sys.argv[1:]):
     with open(main_class_file, 'w') as f:
         f.write(main_class)
 
-
-    launch_java(workspace, jvm_args, main_class, *program_args, additional_jars=args.additional_jars)
+    return launch_java(workspace, jvm_args, main_class, *program_args, additional_jars=args.additional_jars, stdout=stdout, stderr=stderr, check=False)
 
 
 
